unit designer;

{$mode objfpc}

interface

uses
  Classes, SysUtils, libjquery, webwidget, htmlwidgets, contnrs, js, web;

Type
  TIDEClient = Class;

  TIDEResponseHandler = Procedure (aCode : Integer; aCodeText : String; aPayload : TJSObject) of object;

  { TIDERequest }

  TIDERequest = Class(TObject)
  Private
    FXHR : TJSXMLHttpRequest;
    FOnResponse: TIDEResponseHandler;
    Procedure ProcessResponse;
    procedure DoStateChange;
  Public
    Constructor Create(aMethod, aURl: String; aPayLoad: TJSObject; aOnResponse: TIDEResponseHandler);
  end;

  { TIDEClient }
  TCommandEvent = Procedure (Sender : TObject; aCommands : TJSArray) of object;

  TIDEClient = Class(TComponent)
  private
    FPollID : NativeInt;
    FCommandPollInterval : Integer;
    FClientID: NativeInt;
    FIDEURL: String;
    FOnCommands: TCommandEvent;
    FLastPoll : TIDERequest;
    FStartPolling : Boolean;
    procedure DoCommandPoll;
    procedure OnClientRegistered(aCode: Integer; aCodeText: String; aPayload: TJSObject);
    procedure OnCommandsReceived(aCode: Integer; aCodeText: String; aPayload: TJSObject);
  Public
    Constructor Create(aOwner : TComponent); override;
    Procedure RegisterClient;
    Procedure UnRegisterClient;
    Procedure StartCommandPolling;
    Procedure StopCommandPolling;
    Property IDEURL : String read FIDEURL Write FIDEURL;
    Property ClientID : Int64 read FClientID Write FClientID;
    Property CommandPollInterval : Integer Read FCommandPollInterval Write FCommandPollInterval;
    Property OnCommands : TCommandEvent Read FOnCommands Write FOnCommands;
  end;


  { TRegisteredWidget }

  TRegisteredWidget = Class
  Private
    FClass : TWebwidgetClass;
    FImageName : String;
  Public
    Constructor Create(aClass : TWebwidgetClass; aImageName : String);
    Property WidgetClass : TWebwidgetClass Read FClass;
    Property ImageName : String Read FimageName;
  end;

  TWidgetButtonWidget = Class(TButtonWidget)
    MyWidget : TRegisteredWidget;
  end;

  TSortable = Class helper for TJQuery
    Procedure sortable(Options : TJSObject); external name 'sortable'; overload;
    Procedure sortable(Options : string); external name 'sortable'; overload;
  end;
  { TDesignDemo }

  TDesignDemo = class(TComponent)
  private
    FIDEClient: TIDEClient;
    procedure AddWidgetByName(aID: NativeInt; AName: String);
    function CreateNewWidget(aParent: TCustomWebWidget; aClass: TCustomWebWidgetClass): TCustomWebWidget;
    function DoActive(Event: TEventListenerEvent): boolean;
    procedure DoCommandsReceived(Sender: TObject; aCommands: TJSArray);
    procedure DoWidgetAddClick(Sender: TObject; Event: TJSEvent);
    procedure SetIDEClient(AValue: TIDEClient);
    function SortableOptions: TJSObject;
  Public
    FConfirmAdd : NativeInt;
    FAddWidget : TRegisteredWidget;
    FPage : TWebPage;
    FToolBar : TContainerWidget;
    FButtons : Array[1..10] of TButtonWidget;
    FWidgetButtons : Array of TWidgetButtonWidget;
    FRegisteredWidgets : TObjectList;
    Constructor Create(aOwner : TComponent); override;
    Destructor Destroy; override;
    Procedure RegisterWidgets;
    Procedure RegisterWidget(aClass : TWebWidgetClass; aImageName : String);
    procedure SetAddMode(aRegisteredWidget: TRegisteredWidget);
    Procedure FillToolBar;
    Procedure SetupPage;
    property IDEClient: TIDEClient Read FIDEClient Write SetIDEClient;
  end;

implementation

Const
  SSortableSelect = '#designpage, #designpage [data-ww-element-content]';

type
  TWidgetHack = Class(TCustomWebWidget)
    Property Element;
    Property ElementID;
    Property TopElement;
    Property ContentElement;
  end;
  TPageHack  = Class(TWebPage)
    Property Element;
    Property ElementID;
  end;

{ TIDEClient }


procedure TIDEClient.DoCommandPoll;

begin
  if Not Assigned(FLastPoll) then
    FLastPoll:=TIDERequest.Create('Get',IDEURL+'Command/'+IntToStr(ClientID)+'/',Nil,@OnCommandsReceived);
end;

procedure TIDEClient.OnClientRegistered(aCode: Integer; aCodeText: String; aPayload: TJSObject);

begin
  if (aCode div 100)=2 then
    begin
    FClientID:=NativeInt(aPayload['id']);
    if FStartPolling then
      StartCommandPolling;
    end
  else
    FClientID:=0;
end;

procedure TIDEClient.OnCommandsReceived(aCode: Integer; aCodeText: String; aPayload: TJSObject);

Var
  A: TJSArray;

begin
  FLastPoll:=Nil;
  if (aCode div 100)<>2 then
    exit;
  if Assigned(aPayload) and isArray(aPayload['commands']) then
    begin
    A:=TJSArray(aPayload['commands']);
    if (A.Length>0) then
      OnCommands(Self,A);
    end;
end;

constructor TIDEClient.Create(aOwner: TComponent);
begin
  Inherited;
  FLastPoll:=Nil;
  IDEURL:='http://'+Window.location.hostname+':'+Window.location.port+'/IDE/';
end;

procedure TIDEClient.RegisterClient;

Var
  P : TJSObject;
  Req : TIDERequest;

begin
  P:=New(['url',window.locationString]);
  req:=TIDERequest.Create('POST',IDEURL+'Client',P,@OnClientRegistered);
end;

procedure TIDEClient.UnRegisterClient;

Var
  Req : TIDERequest;

begin
  Req:=TIDERequest.Create('DELETE',IDEURL+'Client/'+IntToStr(ClientID),Nil,@OnClientRegistered);
end;

procedure TIDEClient.StartCommandPolling;
begin
  if ClientID<>0 then
    FPollID:=Window.setInterval(@DoCommandPoll,FCommandPollInterval)
  else
    FStartPolling:=True;
end;

procedure TIDEClient.StopCommandPolling;
begin
  FStartPolling:=False;
  if (FPollID>0) then
    Window.clearInterval(FPollID);
end;

{ TIDERequest }

procedure TIDERequest.ProcessResponse;

var
  P : TJSObject;

begin
  if ((FXHR.Status div 100)=2) and (FXHR.ResponseHeaders['Content-Type']='application/json') then
    P:=TJSJSON.parseObject(FXHR.responseText)
  else
    P:=Nil;
  if Assigned(FOnResponse) then
    FOnResponse(FXHR.Status,FXHR.StatusText,P);
end;

procedure TIDERequest.DoStateChange;
begin
  case FXHR.readyState of
    TJSXMLHttpRequest.DONE :
      begin
      if Assigned(FOnResponse) then
        ProcessResponse;
      Free;
      end;
  end;
end;

constructor TIDERequest.Create(aMethod, aURl: String; aPayLoad: TJSObject; aOnResponse: TIDEResponseHandler);

Var
  S : String;

begin
  FOnResponse:=aOnResponse;
  FXHR:=TJSXMLHttpRequest.New;
  FXHR.open(aMethod,aURL);
  if assigned(aPayload) then
    S:=TJSJSON.Stringify(aPayload)
  else
    S:='';
  FXHR.setRequestHeader('Content-Type','application/json');
  FXHR.onreadystatechange:=@DoStateChange;
  FXHR.send(S);
end;

{ TRegisteredWidget }

constructor TRegisteredWidget.Create(aClass: TWebwidgetClass; aImageName: String);
begin
  FClass:=aClass;
  FImageName:=aImageName;
end;

{ TDesignDemo }

function TDesignDemo.CreateNewWidget(aParent : TCustomWebWidget; aClass : TCustomWebWidgetClass) : TCustomWebWidget;

begin
  Result:=aClass.Create(FPage);
  Result.Name:=Result.ClassName+IntToStr(FPage.ChildCount);
  Result.Parent:=aParent;
  Result.Refresh;
end;

function TDesignDemo.DoActive(Event: TEventListenerEvent): boolean;

Const
  Toolbar = '<div class="designerToolbar">' +
            '<div class="designerDragHandle ui-icon ui-icon-arrow-4"></div>' +
            '<div class="designerDelete ui-icon ui-icon-trash"></div>' +
            '</div>';

var
  aParent,aNew : TCustomWebWidget;
  aNewActive : TJSHTMLElement;
  aNewWidget : TRegisteredWidget;

begin
  Result:=True;
  JQuery('.designerActive').removeClass('designerActive');
  JQuery('.designerToolbar').remove();
  aNewActive:=TJSHTMLElement(event.target);
  if (FAddWidget<>Nil) then
    begin
    aNewWidget:=FAddWidget;
    FAddWidget:=Nil;
    aParent:=FPage.FindWidgetByID(String(TJSHTMLElement(event.target).dataset[STopElementData]));
    if aParent<>Nil then
      begin
      aNew:=CreateNewWidget(aParent,aNewWidget.WidgetClass);
      aNewActive:=TWidgetHack(aNew).TopElement;
//      JQuery(SSortableSelect).Sortable('refresh');
      if Assigned(TWidgetHack(aNew).ContentElement) then
         JQuery(TWidgetHack(aNew).ContentElement).Sortable(SortableOptions);
//      JQuery
      jQuery(aNewActive).on_('click',@DoActive);
      end;
    end;
  JQuery(aNewActive).AddClass('designerActive').prepend(toolbar);
end;

procedure TDesignDemo.SetAddMode(aRegisteredWidget : TRegisteredWidget);

begin
  FAddWidget:=aRegisteredWidget;
end;

procedure TDesignDemo.DoWidgetAddClick(Sender: TObject; Event: TJSEvent);
begin
  SetAddMode((Sender as TWidgetButtonWidget).MyWidget);
end;

procedure TDesignDemo.AddWidgetByName(aID : NativeInt; AName : String);

Var
  I : integer;
  Btn : TWidgetButtonWidget;

begin
  I:=FRegisteredWidgets.Count-1;
  While (i>=0) and Not SameText(TRegisteredWidget(FRegisteredWidgets[i]).WidgetClass.ClassName,aName) do
    Dec(I);
  if I<0 then exit;
  SetAddMode(TRegisteredWidget(FRegisteredWidgets[i]));
  FConfirmAdd:=aID;
end;

procedure TDesignDemo.DoCommandsReceived(Sender: TObject; aCommands: TJSArray);

var
  J,P : TJSOBject;
  aName : String;
  aID : NativeInt;
  I : integer;

begin
  for I:=0 to aCommands.Length-1 do
    begin
    J:=TJSObject(aCommands[i]);
    aName:=String(J['name']);
    aID:=NativeInt(J['id']);
    p:=TJSObject(J['payload']);
    case aName of
      'addWidget' : AddWidgetByName(aID,String(P['class']));
    end;
    end;
end;

procedure TDesignDemo.SetIDEClient(AValue: TIDEClient);
begin
  if FIDEClient=AValue then Exit;
  FIDEClient:=AValue;
  if assigned(FIDEClient) then
    begin
    FIDEClient.OnCommands:=@DoCommandsReceived;
    FIDEClient.StartCommandPolling;
    end;
end;

function TDesignDemo.SortableOptions: TJSObject;

begin
  Result:=New([
    'items','> [data-ww-element-top]',
    'connectWith','[data-ww-element-content]',
    'placeholder','designerPlaceholder',
    'tolerance','pointer',
//    'containment',TPageHack(FPage).Element,
//    'handle','[data-ww-element-top]',
    'cancel',''
  ]);
end;

constructor TDesignDemo.Create(aOwner: TComponent);

begin
  inherited Create(aOwner);
  FRegisteredWidgets:=TObjectList.Create;
  RegisterWidgets;
  FillToolBar;
  SetUpPage;
  //  []
  JQuery(SSortableSelect).Sortable(SortableOptions);
  jQuery('#designpage').on_('click','[data-ww-element-top]',@DoActive);
  jQuery('#designpage').on_('click',@DoActive);
end;

destructor TDesignDemo.Destroy;
begin
  FreeAndNil(FRegisteredWidgets);
  inherited Destroy;
end;

procedure TDesignDemo.RegisterWidgets;

begin
  RegisterWidget(TButtonWidget,'button');
  RegisterWidget(TCheckBoxInputWidget,'checkbox');
  RegisterWidget(TRadioInputWidget,'radio');
  RegisterWidget(TTextInputWidget,'edit');
  RegisterWidget(TImageWidget,'image');
  RegisterWidget(TTextAreaWidget,'memo');
  RegisterWidget(TSelectWidget,'select');
  RegisterWidget(TContainerWidget,'container');
end;

procedure TDesignDemo.RegisterWidget(aClass: TWebWidgetClass; aImageName: String);
begin
  FRegisteredWidgets.Add(TRegisteredWidget.Create(aClass,aImageName));
end;

procedure TDesignDemo.SetupPage;

Const
  ButtonClasses : Array[0..8] of string
    = ('primary','secondary','success','danger','warning','info','light','dark','link');

var
  I : Integer;

begin
  FPage:=TWebPage.Create(Self);
  FPage.ElementID:='designpage';
  FPage.Refresh;
  For I:=0 to 9 do
    begin
    FButtons[I]:=TWidgetButtonWidget.Create(FPage);
    FButtons[I].Classes:='btn btn-'+ButtonClasses[I mod 9];
    FButtons[I].Text:='Button #'+IntToStr(I+1);
    FButtons[I].Parent:=FPage;
    FButtons[I].Refresh;
    end;
end;

procedure TDesignDemo.FillToolBar;

Var
  RW : TRegisteredWidget;
  I : Integer;
  Btn : TWidgetButtonWidget;

begin
  FToolBar:=TContainerWidget.Create(Self);
  FToolbar.Styles.EnsureStyle('min-height','34px');
  FToolbar.Styles.RemoveStyle('width');
  FToolbar.Styles.RemoveStyle('height');
  FToolbar.ElementId:='toolbar';
  FToolbar.Refresh;
  SetLength(FWidgetButtons,FRegisteredWidgets.Count);
  For I:=0 to FRegisteredWidgets.Count-1 do
    begin
    RW:=TRegisteredWidget(FRegisteredWidgets[I]);
    Btn:=TWidgetButtonWidget.Create(Self);
    FWidgetButtons[I]:=Btn;
    Btn.MyWidget:=RW;
    Btn.Classes:='btn btn-light';
    Btn.Text:='';
    Btn.Parent:=FToolbar;
    Btn.Styles.Add('background-image','url("widgets/'+RW.ImageName+'.png")');
    Btn.Refresh;
    Btn.Data['widgetClass']:=RW.WidgetClass.ClassName;
    Btn.OnClick:=@DoWidgetAddClick;
    end;
end;

end.

